/*
 * LAMIP-0.0.3 dsp resampler
 *
 * you need  >=libsamplerate-0.1.2
 *
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <plug_dsp.h>

#include <samplerate.h>

static int	src_init(const lamipPluginHandle*);
static void	src_cleanup(void);
static int	src_setup(lamipPCMInfo*);
static int	src_probe(DSPMaterial*);
static int	src_filter(DSPMaterial*);

static DSPPlugin	secret_rabbit_code;
static int		samplerate;
static int		error;
static int		quality;
static SRC_DATA		src_data;
static SRC_STATE*	src_state = NULL;

/****************************************************************************/

DSPPlugin* lamip_dsp_info(void)
{
  secret_rabbit_code.common.name	= "src";
  secret_rabbit_code.common.description = "src - Sample Rate Converter using libsamplerate, the Secret Rabbit Code :)";
  secret_rabbit_code.common.init	= src_init;
  secret_rabbit_code.common.cleanup	= src_cleanup;
  secret_rabbit_code.setup		= src_setup;
  secret_rabbit_code.probe		= src_probe;
  secret_rabbit_code.filter		= src_filter;
  return(&secret_rabbit_code);
}

static int src_init(const lamipPluginHandle* handle)
{
  /* Handling the config options here */
  lamipPluginHandle*	config = (lamipPluginHandle*)handle;
  int			val_samplerate = 0;
  int			val_quality = 0;
	
  if(!lamip_cfg_getExist(handle, "samplerate")){
    lamip_send_message("SRC: src_init: no config value for \"samplerate\" found, resetting to default...\n");
    val_samplerate = 48000;
    lamip_cfg_setInt(config, "samplerate", val_samplerate);
  }else{
    val_samplerate = lamip_cfg_getInt(handle, "samplerate");
  }
  samplerate = val_samplerate;
  if(!lamip_cfg_getExist(handle, "quality")){
    lamip_send_message("SRC: src_init: no config value for \"quality\" found, resetting to default...\n");
    val_quality = 4;
    lamip_cfg_setInt(config, "quality", val_quality);
  }else{
    val_quality = lamip_cfg_getInt(handle, "quality");
  }
  quality = val_quality;
  return(1);
}

static void src_cleanup(void)
{
  /* Normally src_state is still allocated... free it! */
  if(src_state)
    src_delete(src_state);
  return;
}

static int src_setup(lamipPCMInfo* setup)
{
  /* Well... that's simple :) */
  setup->samplerate = samplerate;
  return(1);
}

static int src_probe(DSPMaterial* material)
{
  /* First check if it's 16bit signed little-endian data... */
  if(material->input_pcm_info.format != PCM_FORMAT_S16_LE){
    lamip_send_message("SRC: src_probe: We cannot handle datas other then PCM_FORMAT_S16_LE!\n");
    return(0);
  }
  material->output_pcm_info = material->input_pcm_info;
  /* Check if we need to resample */
  if(material->input_pcm_info.samplerate == samplerate){
    lamip_send_message("SRC: src_probe: We don't need to resample... Already at %lu Hz!\n", samplerate);
    return(0);
  }
  /* Set our samplerate */
  material->output_pcm_info.samplerate = samplerate;
  /* Initializing a SRC_STATE context */
  if(!src_state){
    src_state = src_new(quality, material->input_pcm_info.channels, &error);
    if(!src_state){
      fprintf(stderr, "SRC: src_probe: src_new() failed...\n");
      return(0);
    }
    fprintf(stderr, "SRC: src_probe: src_state initialized!\n");
  }
  /* Initializing some things for the SRC_DATA thing */
  src_data.src_ratio = (double)((double)samplerate / (double)material->input_pcm_info.samplerate);
  src_data.end_of_input = 0;
  return(1);
}

static int src_filter(DSPMaterial* material)
{
  int i;
  int length = 0;
  int size;
  int bound;
  int output_len;
  /* pointers to the buffers necessary for the conversion between float and integer datas */
  int16_t*	int16;
  float*	float32;
	
  /* Input buffer must be dividable by 4 (sizeof int16 and 2 channels) */
  if( material->input_buffer_len % 4 != 0){
    lamip_send_message("SRC: src_init: F...! the internal buffer len isn't dividable by 4! NOTHING will work! Aborting...\n");
    return(0);
  }
  /* Computing some important variables */
  length = (int)(material->input_buffer_len / 4);  /* 2chan multiplied by int16 = 4 */
  bound = (int) ceil((src_data.src_ratio + 0.05) * length); 
  /* Allocating SRC_DATA.data_in buffer */
  size = length * 2 * sizeof(float);
  src_data.data_in = malloc(size);
  if(!src_data.data_in){
    fprintf(stderr, "malloc failed!\n");
    return(0);
  }
  src_data.input_frames = length;
  /* Allocating SRC_DATA.data_out buffer */
  size = bound * 2 * sizeof(float);
  src_data.data_out = malloc(size);
  if(!src_data.data_out){
    fprintf(stderr, "malloc failed!\n");
    free(src_data.data_in);
    return(0);
  }
  src_data.output_frames = bound;
  /* Doing the int16 to float conversion : */
  float32 = src_data.data_in;
  int16 = (int16_t *)material->input_buffer;
  /* Just for interest: The "else" part does the same as the 2 following lines
   * But it's based on some trick of the floating point format...
   * the conversion 
   * is much faster like that!
   */
  /* Slow version : */
#if 0
  for(i=0; i < 2*length; i++)
    *(float32++) = (float) *(int16++) / 32768.;
#else
  /* Fast version : learned that while reading through the VLC sources!
   * Great work guys!
   */
  for(i=0; i < 2*length; i++) {
    union { float g; int32_t i; } u;
    u.i = *int16 + 0x43c00000;
    *float32 = u.g - 384.0;
    float32++;
    int16++;
  }
#endif
  /* Checking if the conversion was succesfull! */
  if(float32 != (src_data.data_in + length*2)){
    fprintf(stderr, "SRC: src_filter: conversion wasn't succesfull!\n");
    free(src_data.data_in);
    free(src_data.data_out);
    return(0);
  }
  /***********************************
   * MAIN thing - the resampling! :) *
   ***********************************/
  error = src_process(src_state, & src_data);
  if(error){
    fprintf(stderr, "src_process returned %d: %s\n", error, src_strerror(error));
  }
  /* resampling done... checking if everything is ok! */
  output_len = src_data.output_frames_gen;
  if(!(output_len <= bound)){
    fprintf(stderr, "output_len <= bound!\n");
    free(src_data.data_in);
    free(src_data.data_out);
    return(0);
  }
  if(src_data.input_frames_used != length) {
    lamip_send_message("SRC: src_filter: SRC didn't use all frames! Aborting...\n");
    free(src_data.data_in);
    free(src_data.data_out);
    return(0);
  }
  /* Allocating DSPMaterial output buffer! */
  size = bound * 2 * sizeof(int16_t);
  material->output_buffer = malloc(size);
  if(!material->output_buffer){
    fprintf(stderr, "malloc failed!\n");
    free(src_data.data_in);
    free(src_data.data_out);
    return(0);
  }
  /* now converting back to material->output_pcm_info.format --> int16 at the moment only */
  float32 = src_data.data_out;
  int16 = material->output_buffer;
  material->output_buffer_len = output_len * 4;
  for(i=0; i<output_len; i++){
    float v_l = 32768* *(float32 ++);
    float v_r = 32768* *(float32 ++);
    /* TODO: better anti-clipping */
    if(v_l > 32767.0)
      v_l = 32767.0;
    else if(v_l < -32767.0)
      v_l = -32767.0;
    if(v_r > 32767.0)
      v_r = 32767.0;
    else if(v_r < -32767.0)
      v_r = -32767.0;
    *(int16 ++) = v_l;
    *(int16 ++) = v_r;
  }
#if 0	
  for(i=0; i<output_len; i++) {
    union { float f; int32_t i; } u;
    u.f = *float32 + 384.0;
    if ( u.i > 0x43c07fff ) *int16 = 32767;
    else if ( u.i < 0x43bf8000 ) *int16 = -32768;
    else *int16 = u.i - 0x43c00000;
    int16++;
    float32++;
  }
#endif
  /* finally! freeing our buffers for conversion ... */
  free(src_data.data_in);
  free(src_data.data_out);
  return(1);
}
